package spacedata

import (
	"database/sql"
	"errors"
	"fmt"
	"log"

	_ "github.com/mattn/go-sqlite3"
)

const initialVersion int64 = 1

const currentVersion = initialVersion

type PeopleStorage struct {
	DatabaseName     string
	DatabasePath     string
	databaseFilePath string
}

type PeopleDataSet struct {
	Name         string
	Craft        string
	firstIsoDate string
	lastIsoDate  string
}

func (s *PeopleStorage) checkDatabase() error {
	if s.databaseFilePath == "" {
		name := s.DatabasePath + s.DatabaseName
		if name == "" {
			name = "./spacecowboys.db"
		}
		s.databaseFilePath = name
	}

	db, err := sql.Open("sqlite3", s.databaseFilePath)
	if err != nil {
		log.Fatalf("Unable to open database %#v db: error: %v", s.databaseFilePath, err)
	}
	defer db.Close()

	creationString := "CREATE TABLE IF NOT EXISTS spaceguys (id INTEGER NOT NULL PRIMARY KEY, name TEXT, craft TEXT, firstAppearance DATETIME, lastAppearance DATETIME); CREATE UNIQUE INDEX IF NOT EXISTS guySpaceIndex ON spaceguys(name, craft);"

	_, err = db.Exec(creationString)
	if err != nil {
		log.Fatalf("Unable to create table spaceguys in Database %#v at path %#v: error: %v", db, s.databaseFilePath, err)
	}

	creationString = "CREATE TABLE IF NOT EXISTS version (id INTEGER NOT NULL PRIMARY KEY, version INTEGER);"

	_, err = db.Exec(creationString)
	if err != nil {
		log.Fatalf("Unable to create table spaceguys in Database %#v at path %#v: error: %v", db, s.databaseFilePath, err)
	}

	queryString := "SELECT version FROM version ORDER BY id DESC LIMIT 1"
	rows, err := db.Query(queryString)
	if err != nil {
		return err
	}
	defer rows.Close()

	if rows.Next() {
		var latestVersion int64
		err = rows.Scan(&latestVersion)
		if err != nil {
			return err
		}

		if latestVersion < currentVersion {
			// DO SOME MIGRATION!

			// Raise version information.
			res, err := db.Exec("INSERT INTO version (version) VALUES (?)", currentVersion)
			if err != nil {
				return err
			}
			if res == nil {
				return errors.New("unable to create first version information")
			}
		}
	} else {
		res, err := db.Exec("INSERT INTO version (version) VALUES (?)", currentVersion)
		if err != nil {
			return err
		}
		if res == nil {
			return errors.New("unable to create first version information")
		}
	}
	return nil
}

func (s *PeopleStorage) Post(d *PeopleDataSet) (rowIndex int64, sqliteError error) {
	rowIndex = 0
	sqliteError = s.checkDatabase()
	if sqliteError != nil {
		return
	}

	db, sqliteError := sql.Open("sqlite3", s.databaseFilePath)
	if sqliteError != nil {
		return
	}
	defer db.Close()

	transaction, sqliteError := db.Begin()
	if sqliteError != nil {
		return
	}

	queryString := "SELECT id FROM spaceguys WHERE name LIKE ? AND craft LIKE ?"
	result, sqliteError := transaction.Query(queryString, d.Name, d.Craft)
	if sqliteError != nil {
		return
	}
	if result == nil {
		reason := fmt.Sprintf("Got nil result for query %#v with name %#v and craft %#v", queryString, d.Name, d.Craft)
		sqliteError = errors.New(reason)
		return
	}

	if result.Next() {
		var foundRow int64
		sqliteError = result.Scan(&foundRow)
		if sqliteError != nil {
			return
		}
		queryString = "UPDATE spaceguys SET name = ?, craft = ?, lastAppearance = strftime('%Y-%m-%dT%H:%M:%fZ', 'now') WHERE id = " + fmt.Sprintf("%v", foundRow)
	} else {
		queryString = "INSERT INTO spaceguys (name, craft, firstAppearance, lastAppearance) VALUES (?, ?, strftime('%Y-%m-%dT%H:%M:%fZ', 'now'), strftime('%Y-%m-%dT%H:%M:%fZ', 'now'))"
	}

	sqlStatement, sqliteError := transaction.Prepare(queryString)
	if sqliteError != nil {
		return
	}
	defer sqlStatement.Close()

	res, sqliteError := sqlStatement.Exec(d.Name, d.Craft)
	if sqliteError != nil {
		return
	}

	if res == nil {
		reason := fmt.Sprintf("Table creation for spaceguys did not return resource information in database %#v", db)
		sqliteError = errors.New(reason)
		return
	}
	row, sqliteError := res.LastInsertId()
	if sqliteError != nil {
		return
	}
	affected, sqliteError := res.RowsAffected()
	if sqliteError != nil {
		return
	}

	if row == 0 && affected == 0 {
		reason := fmt.Sprintf("Table creation/update returned rowId and affected row amount of 0 which is ridiculous")
		sqliteError = errors.New(reason)
		return
	}
	sqliteError = transaction.Commit()

	rowIndex = row
	if rowIndex == 0 {
		rowIndex = affected
	}

	return
}

func (s *PeopleStorage) Get(row int64) (*PeopleDataSet, error) {
	err := s.checkDatabase()
	if err != nil {
		return nil, err
	}

	db, err := sql.Open("sqlite3", s.databaseFilePath)
	if err != nil {
		return nil, err
	}
	defer db.Close()

	queryString := "SELECT name, craft, firstAppearance, lastAppearance FROM spaceguys WHERE id = ?"
	rows, err := db.Query(queryString, row)
	if err != nil {
		return nil, err
	}

	if rows == nil {
		reason := fmt.Sprintf("Got nil result for query %#v with name %#v", queryString, row)
		queryError := errors.New(reason)
		return nil, queryError
	}
	defer rows.Close()

	// Since we select by id we assume there is at most one result to store.
	loadedGuys := make([]PeopleDataSet, 0)
	for rows.Next() {
		var name string
		var craft string
		var first string
		var last string
		err = rows.Scan(&name, &craft, &first, &last)
		if err != nil {
			return nil, err
		}
		loadedGuy := PeopleDataSet{Name: name, Craft: craft, firstIsoDate: first, lastIsoDate: last}
		loadedGuys = append(loadedGuys, loadedGuy)
	}

	if len(loadedGuys) == 0 {
		return nil, nil
	}

	result := loadedGuys[0]
	if len(loadedGuys) > 1 {
		reason := fmt.Sprintf("select statement for row %#v returned %v results, which seems ridicolous. Result is first occurance.", row, len(loadedGuys))
		err = errors.New(reason)
	}

	return &result, nil
}

func (s *PeopleStorage) GetPassengers(craft string) ([]PeopleDataSet, error) {
	err := s.checkDatabase()
	if err != nil {
		return nil, err
	}

	db, err := sql.Open("sqlite3", s.databaseFilePath)
	if err != nil {
		return nil, err
	}
	defer db.Close()

	rows, err := db.Query("SELECT name FROM spaceguys WHERE craft LIKE ?", craft)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	result := make([]PeopleDataSet, 0)
	for rows.Next() {
		var name string
		err = rows.Scan(&name)
		guy := PeopleDataSet{Name: name, Craft: craft}
		result = append(result, guy)
	}
	return result, err
}
