package spacedata

import (
	"os"
	"testing"
)

func prepareDatabase(databaseFilePath string) error {
	err := os.Remove(databaseFilePath)
	if !os.IsNotExist(err) {
		return err
	}
	return nil
}

func TestPeopleStorage_Post(t *testing.T) {
	dbPath := "./"
	dbName := "singleguy.db"
	err := prepareDatabase(dbPath + dbName)
	if err != nil {
		t.Fatalf("Unable to prepare database at path %#v; got error: %v", dbPath+dbName, err)
	}
	guyInTesting := PeopleDataSet{Name: "Laika", Craft: "Sputnik 2"}
	storageInTesting := PeopleStorage{DatabaseName: dbName, DatabasePath: dbPath}

	row, err := storageInTesting.Post(&guyInTesting)

	if err != nil {
		t.Errorf("creation of %#v returned error: %v", guyInTesting, err)
	}

	if row == 0 {
		t.Errorf("creation of %#v did return 0 as rowId which is impossibru", guyInTesting)
	}

	info, err := os.Stat(dbPath + dbName)
	if err != nil {
		t.Fatalf("Invalid database filename %#v at path %#v: getting info returned error: %v", dbName, dbPath, err)
	}
	if info.Name() != dbName {
		t.Errorf("db name mismatch: wanted %#v; got %#v", dbName, info.Name())
	}

	guyFromDatabase, err := storageInTesting.Get(row)

	if err != nil {
		t.Errorf("reading created entry at row %#v failed with error: %v", row, err)
	}
	if guyFromDatabase == nil {
		t.Fatalf("Got no guy result from database. Aborting to avoid panic.")
	}

	if guyFromDatabase.Name != guyInTesting.Name {
		t.Errorf("read guys name mismatch: got %#v; wanted %#v", guyFromDatabase.Name, guyInTesting.Name)
	}
	if guyFromDatabase.Craft != guyInTesting.Craft {
		t.Errorf("read guys ship mismatch: got %#v; wanted %#v", guyFromDatabase.Craft, guyInTesting.Craft)
	}

	firstIsoDate := guyFromDatabase.firstIsoDate
	lastIsoDate := guyFromDatabase.lastIsoDate

	newRow, err := storageInTesting.Post(&guyInTesting)
	if err != nil {
		t.Errorf("posting again returned error: %v", err)
	}
	if row != newRow {
		t.Errorf("old row was %#v, new row is %#v for the same thing", row, newRow)
	}

	guyFromDatabase, err = storageInTesting.Get(newRow)
	if err != nil {
		t.Errorf("posting again returned error: %v", err)
	}
	if guyFromDatabase == nil {
		t.Fatalf("got nil result reading guys from database. Aborting since we may panic.")
	}
	if guyFromDatabase.Name != guyInTesting.Name {
		t.Errorf("read guys name mismatch: got %#v; wanted %#v", guyFromDatabase.Name, guyInTesting.Name)
	}
	if guyFromDatabase.Craft != guyInTesting.Craft {
		t.Errorf("read guys ship mismatch: got %#v; wanted %#v", guyFromDatabase.Craft, guyInTesting.Craft)
	}

	if guyFromDatabase.firstIsoDate != firstIsoDate {
		t.Errorf("expected first iso date to be equal, but old was %#v and new is %#v", firstIsoDate, guyFromDatabase.firstIsoDate)
	}
	if guyFromDatabase.lastIsoDate == lastIsoDate {
		t.Errorf("there should have been an update between both last dates")
	}
}

func TestPeopleStorage_Get(t *testing.T) {
	dbPath := "./"
	dbName := "moreGuys.db"
	err := prepareDatabase(dbPath + dbName)
	if err != nil {
		t.Fatalf("Unable to prepare database at path %#v; got error: %v", dbPath+dbName, err)
	}
	storageInTesting := PeopleStorage{DatabaseName: dbName, DatabasePath: dbPath}
	spacecraft := "Saturn V"

	buzz := PeopleDataSet{Name: "Buzz Aldrin", Craft: spacecraft}
	michael := PeopleDataSet{Name: "Michael Collins", Craft: spacecraft}
	neil := PeopleDataSet{Name: "Neil Armstrong", Craft: spacecraft}
	_, _ = storageInTesting.Post(&michael)
	_, _ = storageInTesting.Post(&neil)
	_, _ = storageInTesting.Post(&buzz)

	guys, err := storageInTesting.GetPassengers(spacecraft)

	if err != nil {
		t.Errorf("getting all saturn v guys failed with error: %v", err)
	}
	if len(guys) != 3 {
		t.Fatalf("amount of resulting passengers mismatch: got %d; wanted %d", len(guys), 3)
	}
	for _, guy := range guys {
		if guy.Name != buzz.Name && guy.Name != michael.Name && guy.Name != neil.Name {
			t.Errorf("guy name mismatch: unknown passenger %#v for craft %#v", guy.Name, spacecraft)
		}
	}
}
